@@ -863,7 +863,14 @@ int Client::on_read(const Endpoint &ep) {
     .msg_control = msg_ctrl,
   };
 
-  for (;;) {
+  auto start = util::timestamp();
+
+  for (; pktcnt < MAX_RECV_PKTS;) {
+    if (util::recv_pkt_time_threshold_exceeded(
+          config.cc_algo == NGTCP2_CC_ALGO_BBR, start, pktcnt)) {
+      break;
+    }
+
     msg.msg_namelen = sizeof(su);
     msg.msg_controllen = sizeof(msg_ctrl);
 
@@ -924,10 +931,6 @@ int Client::on_read(const Endpoint &ep) {
         break;
       }
     }
-
-    if (pktcnt >= 10) {
-      break;
-    }
   }
 
   if (should_exit()) {
@@ -1058,17 +1061,20 @@ int Client::write_streams() {
   size_t gso_size;
   auto ts = util::timestamp();
   auto txbuf = std::span{tx_.data};
+  auto buflen = util::clamp_buffer_size(conn_, txbuf.size(), config.gso_burst);
 
   ngtcp2_path_storage_zero(&ps);
 
-  auto nwrite =
-    ngtcp2_conn_write_aggregate_pkt(conn_, &ps.path, &pi, txbuf.data(),
-                                    txbuf.size(), &gso_size, ::write_pkt, ts);
+  auto nwrite = ngtcp2_conn_write_aggregate_pkt2(
+    conn_, &ps.path, &pi, txbuf.data(), buflen, &gso_size, ::write_pkt,
+    config.gso_burst, ts);
   if (nwrite < 0) {
     disconnect();
     return -1;
   }
 
+  ngtcp2_conn_update_pkt_tx_time(conn_, ts);
+
   if (nwrite == 0) {
     return 0;
   }
@@ -2189,6 +2195,12 @@ Options:
   --no-gso    Disables GSO.
   --show-stat Print the connection statistics when the connection is
               closed.
+  --gso-burst=<N>
+              The maximum number of packets  to aggregate for GSO.  If
+              GSO is disabled,  this is the maximum  number of packets
+              to send  per an event  loop in a single  connection.  It
+              defaults  to 0,  which means  it is  not limited  by the
+              configuration.
   -h, --help  Display this help and exit.
 
 ---
@@ -2275,6 +2287,7 @@ int main(int argc, char **argv) {
       {"ech-config-list-file", required_argument, &flag, 44},
       {"no-gso", no_argument, &flag, 45},
       {"show-stat", no_argument, &flag, 46},
+      {"gso-burst", required_argument, &flag, 47},
       {},
     };
 
@@ -2585,9 +2598,9 @@ int main(int argc, char **argv) {
         if (auto n = util::parse_uint_iec(optarg); !n) {
           std::cerr << "max-udp-payload-size: invalid argument" << std::endl;
           exit(EXIT_FAILURE);
-        } else if (*n > 64_k) {
-          std::cerr << "max-udp-payload-size: must not exceed 65536"
-                    << std::endl;
+        } else if (*n > NGTCP2_MAX_TX_UDP_PAYLOAD_SIZE) {
+          std::cerr << "max-udp-payload-size: must not exceed "
+                    << NGTCP2_MAX_TX_UDP_PAYLOAD_SIZE << std::endl;
           exit(EXIT_FAILURE);
         } else if (*n == 0) {
           std::cerr << "max-udp-payload-size: must not be 0" << std::endl;
@@ -2705,10 +2718,12 @@ int main(int argc, char **argv) {
           if (auto n = util::parse_uint_iec(s); !n) {
             std::cerr << "pmtud-probes: invalid argument" << std::endl;
             exit(EXIT_FAILURE);
-          } else if (*n <= 1200 || *n >= 64_k) {
-            std::cerr
-              << "pmtud-probes: must be in range [1201, 65535], inclusive."
-              << std::endl;
+          } else if (*n <= NGTCP2_MAX_UDP_PAYLOAD_SIZE ||
+                     *n > NGTCP2_MAX_TX_UDP_PAYLOAD_SIZE) {
+            std::cerr << "pmtud-probes: must be in range ["
+                      << NGTCP2_MAX_UDP_PAYLOAD_SIZE + 1 << ", "
+                      << NGTCP2_MAX_TX_UDP_PAYLOAD_SIZE << "], inclusive."
+                      << std::endl;
             exit(EXIT_FAILURE);
           } else {
             config.pmtud_probes.push_back(static_cast<uint16_t>(*n));
@@ -2728,11 +2743,29 @@ int main(int argc, char **argv) {
         // --show-stat
         config.show_stat = true;
         break;
+      case 47: {
+        // --gso-burst
+        auto n = util::parse_uint(optarg);
+        if (!n) {
+          std::cerr << "gso-burst: invalid argument" << std::endl;
+          exit(EXIT_FAILURE);
+        }
+
+        if (*n > 64) {
+          std::cerr << "gso-burst: must be in range [0, 64], inclusive."
+                    << std::endl;
+          exit(EXIT_FAILURE);
+        }
+
+        config.gso_burst = static_cast<size_t>(*n);
+
+        break;
+      }
       }
       break;
     default:
       break;
-    };
+    }
   }
 
   if (argc - optind < 2) {
