@@ -29,13 +29,13 @@
 #  include <config.h>
 #endif // defined(HAVE_CONFIG_H)
 
-#include <string>
+#include <string_view>
 
 namespace ngtcp2 {
 
 namespace http {
 
-std::string get_reason_phrase(unsigned int status_code);
+std::string_view get_reason_phrase(unsigned int status_code);
 
 } // namespace http
 
