@@ -27,7 +27,7 @@
 
 #ifdef HAVE_CONFIG_H
 #  include <config.h>
-#endif // HAVE_CONFIG_H
+#endif // defined(HAVE_CONFIG_H)
 
 #include <string>
 #include <string_view>
@@ -51,4 +51,4 @@ protected:
   SSL *ssl_;
 };
 
-#endif // TLS_SESSION_BASE_QUICTLS_H
+#endif // !defined(TLS_SESSION_BASE_QUICTLS_H)
